@@ -23,11 +23,11 @@ from mercurial import (
     context,
     encoding,
     util as hgutil,
-    lock as lockmod,
     url,
 )
 
 import _ssh
+import compat
 import git2hg
 import hg2git
 import util
@@ -54,6 +54,8 @@ RE_GIT_PROGRESS = re.compile('\((\d+)/(\d+)\)')
 
 RE_AUTHOR_FILE = re.compile('\s*=\s*')
 
+CALLBACK_BUFFER = ''
+
 class GitProgress(object):
     """convert git server progress strings into mercurial progress"""
     def __init__(self, ui):
@@ -104,9 +106,9 @@ class GitHandler(object):
         self.ui = ui
 
         if ui.configbool('git', 'intree'):
-            self.gitdir = self.repo.wjoin('.git')
+            self.gitdir = self.repo.wvfs.join('.git')
         else:
-            self.gitdir = self.repo.join('git')
+            self.gitdir = self.repo.vfs.join('git')
 
         self.init_author_file()
 
@@ -157,7 +159,7 @@ class GitHandler(object):
     def init_author_file(self):
         self.author_map = {}
         if self.ui.config('git', 'authors'):
-            with open(self.repo.wjoin(self.ui.config('git', 'authors'))) as f:
+            with open(self.repo.wvfs.join(self.ui.config('git', 'authors'))) as f:
                 for line in f:
                     line = line.strip()
                     if not line or line.startswith('#'):
@@ -180,8 +182,8 @@ class GitHandler(object):
     def load_map(self):
         map_git_real = {}
         map_hg_real = {}
-        if os.path.exists(self.repo.join(self.map_file)):
-            for line in self.repo.opener(self.map_file):
+        if os.path.exists(self.repo.vfs.join(self.map_file)):
+            for line in self.repo.vfs(self.map_file):
                 # format is <40 hex digits> <40 hex digits>\n
                 if len(line) != 82:
                     raise ValueError(
@@ -195,26 +197,30 @@ class GitHandler(object):
         self._map_hg_real = map_hg_real
 
     def save_map(self, map_file):
-        file = self.repo.opener(map_file, 'w+', atomictemp=True)
-        map_hg = self._map_hg
-        buf = cStringIO.StringIO()
-        bwrite = buf.write
-        for hgsha, gitsha in map_hg.iteritems():
-            bwrite("%s %s\n" % (gitsha, hgsha))
-        file.write(buf.getvalue())
-        buf.close()
-        # If this complains, atomictempfile no longer has close
-        file.close()
+        wlock = self.repo.wlock()
+        try:
+            file = self.repo.vfs(map_file, 'w+', atomictemp=True)
+            map_hg = self._map_hg
+            buf = cStringIO.StringIO()
+            bwrite = buf.write
+            for hgsha, gitsha in map_hg.iteritems():
+                bwrite("%s %s\n" % (gitsha, hgsha))
+            file.write(buf.getvalue())
+            buf.close()
+            # If this complains, atomictempfile no longer has close
+            file.close()
+        finally:
+            wlock.release()
 
     def load_tags(self):
         self.tags = {}
-        if os.path.exists(self.repo.join(self.tags_file)):
-            for line in self.repo.opener(self.tags_file):
+        if os.path.exists(self.repo.vfs.join(self.tags_file)):
+            for line in self.repo.vfs(self.tags_file):
                 sha, name = line.strip().split(' ', 1)
                 self.tags[name] = sha
 
     def save_tags(self):
-        file = self.repo.opener(self.tags_file, 'w+', atomictemp=True)
+        file = self.repo.vfs(self.tags_file, 'w+', atomictemp=True)
         for name, sha in sorted(self.tags.iteritems()):
             if not self.repo.tagtype(name) == 'global':
                 file.write("%s %s\n" % (sha, name))
@@ -259,6 +265,9 @@ class GitHandler(object):
         refs = self.fetch_pack(remote, heads)
         remote_name = self.remote_name(remote)
 
+        # if remote returns a symref for HEAD, then let's store that
+        rhead = None
+        rnode = None
         oldheads = self.repo.changelog.heads()
         imported = 0
         if refs:
@@ -266,6 +275,20 @@ class GitHandler(object):
             imported = self.import_git_objects(remote_name, filteredrefs)
             self.import_tags(refs)
             self.update_hg_bookmarks(refs)
+
+            try:
+                symref = refs['HEAD']
+                if symref.startswith('refs/heads'):
+                    rhead = symref.replace('refs/heads/', '')
+
+                rnode = refs['refs/heads/%s' % rhead]
+                rnode = self._map_git[rnode]
+                rnode = self.repo[rnode].node()
+            except KeyError:
+                # if there is any error make sure to clear the variables
+                rhead = None
+                rnode = None
+
             if remote_name:
                 self.update_remote_branches(remote_name, refs)
             elif not self.paths:
@@ -275,6 +298,15 @@ class GitHandler(object):
                 # "Activate" a tipmost bookmark.
                 bms = self.repo['tip'].bookmarks()
 
+                # override the 'tipmost' behavior if we know the remote HEAD
+                if rnode:
+                    # make sure the bookmark exists; at the point the remote
+                    # branches has already been set up
+                    suffix = self.branch_bookmark_suffix or ''
+                    changes = [(rhead + suffix, rnode)]
+                    util.updatebookmarks(self.repo, changes)
+                    bms = [rhead + suffix]
+
                 if bms:
                     try:
                         bookmarks.activate(self.repo, bms[0])
@@ -374,7 +406,7 @@ class GitHandler(object):
         return ret
 
     def clear(self):
-        mapfile = self.repo.join(self.map_file)
+        mapfile = self.repo.vfs.join(self.map_file)
         if os.path.exists(self.gitdir):
             for root, dirs, files in os.walk(self.gitdir, topdown=False):
                 for name in files:
@@ -409,13 +441,27 @@ class GitHandler(object):
     # CHANGESET CONVERSION METHODS
 
     def export_git_objects(self):
+        self.ui.note(_("finding hg commits to export\n"))
         repo = self.repo
         clnode = repo.changelog.node
+
         nodes = (clnode(n) for n in repo)
-        export = (repo[node] for node in nodes if not hex(node) in
+        to_export = (repo[node] for node in nodes if not hex(node) in
                   self._map_hg)
-        export = [ctx for ctx in export
-                  if ctx.extra().get('hg-git', None) != 'octopus']
+
+        todo_total = len(repo) - len(self._map_hg)
+        topic = 'find commits to export'
+        pos = 0
+        unit = 'commits'
+
+        export = []
+        for ctx in to_export:
+            item = hex(ctx.node())
+            pos += 1
+            repo.ui.progress(topic, pos, item, unit, todo_total)
+            if ctx.extra().get('hg-git', None) != 'octopus':
+                export.append(ctx)
+
         total = len(export)
         if not total:
             return
@@ -443,11 +489,20 @@ class GitHandler(object):
         exporter = hg2git.IncrementalChangesetExporter(
             self.repo, pctx, self.git.object_store, gitcommit)
 
+        mapsavefreq = self.ui.configint('hggit', 'mapsavefrequency', 0)
         for i, ctx in enumerate(export):
             self.ui.progress('exporting', i, total=total)
             self.export_hg_commit(ctx.node(), exporter)
+            if mapsavefreq and i % mapsavefreq == 0:
+                self.ui.debug(_("saving mapfile\n"))
+                self.save_map(self.map_file)
         self.ui.progress('exporting', None, total=total)
 
+    def set_commiter_from_author(self, commit):
+        commit.committer = commit.author
+        commit.commit_time = commit.author_time
+        commit.commit_timezone = commit.author_timezone
+
     # convert this commit into git objects
     # go through the manifest, convert all blobs/trees we don't have
     # write the commit object (with metadata info)
@@ -473,25 +528,26 @@ class GitHandler(object):
         commit.author_timezone = -timezone
 
         if 'committer' in extra:
-            # fixup timezone
-            (name, timestamp, timezone) = extra['committer'].rsplit(' ', 2)
-            commit.committer = name
-            commit.commit_time = timestamp
+            try:
+                # fixup timezone
+                (name, timestamp, timezone) = extra['committer'].rsplit(' ', 2)
+                commit.committer = name
+                commit.commit_time = timestamp
 
-            # work around a timezone format change
-            if int(timezone) % 60 != 0:  # pragma: no cover
-                timezone = parse_timezone(timezone)
-                # Newer versions of Dulwich return a tuple here
-                if isinstance(timezone, tuple):
-                    timezone, neg_utc = timezone
-                    commit._commit_timezone_neg_utc = neg_utc
-            else:
-                timezone = -int(timezone)
-            commit.commit_timezone = timezone
+                # work around a timezone format change
+                if int(timezone) % 60 != 0:  # pragma: no cover
+                    timezone = parse_timezone(timezone)
+                    # Newer versions of Dulwich return a tuple here
+                    if isinstance(timezone, tuple):
+                        timezone, neg_utc = timezone
+                        commit._commit_timezone_neg_utc = neg_utc
+                else:
+                    timezone = -int(timezone)
+                commit.commit_timezone = timezone
+            except: # extra is essentially user-supplied, we must be careful
+                self.set_commiter_from_author(commit)
         else:
-            commit.committer = commit.author
-            commit.commit_time = commit.author_time
-            commit.commit_timezone = commit.author_timezone
+            self.set_commiter_from_author(commit)
 
         commit.parents = []
         for parent in self.get_git_parents(ctx):
@@ -560,8 +616,11 @@ class GitHandler(object):
 
         TESTS:
 
+        >>> from collections import namedtuple
         >>> from mercurial.ui import ui
-        >>> g = GitHandler('', ui()).get_valid_git_username_email
+        >>> mockrepo = namedtuple('localrepo', ['vfs'])
+        >>> mockrepo.vfs = ''
+        >>> g = GitHandler(mockrepo, ui()).get_valid_git_username_email
         >>> g('John Doe')
         'John Doe'
         >>> g('john@doe.com')
@@ -929,6 +988,8 @@ class GitHandler(object):
         # if named branch, add to extra
         if hg_branch:
             extra['branch'] = hg_branch
+        else:
+            extra['branch'] = 'default'
 
         # if committer is different than author, add it to extra
         if commit.author != commit.committer \
@@ -980,7 +1041,7 @@ class GitHandler(object):
                 for rev in (hex(r) for r in revs):
                     if rev not in all_exportable:
                         raise hgutil.Abort("revision %s cannot be pushed since"
-                                           " it doesn't have a ref" %
+                                           " it doesn't have a bookmark" %
                                            self.repo[rev])
                     exportable[rev] = all_exportable[rev]
             return self.get_changed_refs(refs, exportable, force)
@@ -1004,9 +1065,19 @@ class GitHandler(object):
             return self.git.object_store.generate_pack_contents(have, want)
 
         def callback(remote_info):
+            # dulwich (perhaps git?) wraps remote output at a fixed width but
+            # signifies the end of transmission with a double new line
+            global CALLBACK_BUFFER
+            if remote_info and not remote_info.endswith('\n\n'):
+                CALLBACK_BUFFER += remote_info
+                return
+
+            remote_info = CALLBACK_BUFFER + remote_info
+            CALLBACK_BUFFER = ''
             if not remote_info:
-                remote_info = ''
-            for line in remote_info.split('\n'):
+                remote_info = '\n'
+
+            for line in remote_info[:-1].split('\n'):
                 self.ui.status(_("remote: %s\n") % line)
 
         try:
@@ -1058,7 +1129,7 @@ class GitHandler(object):
             ctx = self.repo[rev]
             if not rev_refs:
                 raise hgutil.Abort("revision %s cannot be pushed since"
-                                   " it doesn't have a ref" % ctx)
+                                   " it doesn't have a bookmark" % ctx)
 
             # Check if the tags the server is advertising are annotated tags,
             # by attempting to retrieve it from the our git repo, and building
@@ -1124,6 +1195,10 @@ class GitHandler(object):
         try:
             progress = GitProgress(self.ui)
             f = StringIO.StringIO()
+
+            # monkey patch dulwich's read_pkt_refs so that we can determine on
+            # clone which bookmark to activate
+            client.read_pkt_refs = compat.read_pkt_refs
             ret = localclient.fetch_pack(path, determine_wants, graphwalker,
                                          f.write, progress.progress)
             if(f.pos != 0):
@@ -1303,6 +1378,7 @@ class GitHandler(object):
                           if ref.startswith('refs/heads/')])
 
             suffix = self.branch_bookmark_suffix or ''
+            changes = []
             for head, sha in heads.iteritems():
                 # refs contains all the refs in the server, not just
                 # the ones we are pulling
@@ -1312,27 +1388,15 @@ class GitHandler(object):
                 hgsha = bin(hgsha)
                 if head not in bms:
                     # new branch
-                    bms[head + suffix] = hgsha
+                    changes.append((head + suffix, hgsha))
                 else:
                     bm = self.repo[bms[head]]
                     if bm.ancestor(self.repo[hgsha]) == bm:
                         # fast forward
-                        bms[head + suffix] = hgsha
+                        changes.append((head + suffix, hgsha))
 
             if heads:
-                tr = lock = wlock = None
-                try:
-                    wlock = self.repo.wlock()
-                    lock = self.repo.lock()
-                    tr = self.repo.transaction('git_handler')
-                    if hgutil.safehasattr(bms, 'recordchange'):
-                        # recordchange was added in mercurial 3.2
-                        bms.recordchange(tr)
-                    else:
-                        bms.write()
-                    tr.close()
-                finally:
-                    lockmod.release(tr, lock, wlock)
+                util.updatebookmarks(self.repo, changes)
 
         except AttributeError:
             self.ui.warn(_('creating bookmarks failed, do you have'
@@ -1457,6 +1521,7 @@ class GitHandler(object):
                 gitlinks[oldfile] = None
                 continue
             if newfile is not None:
+                self.audit_hg_path(newfile)
                 # new = file
                 files[newfile] = False, newmode, newsha
                 if renames is not None and newfile != oldfile:
@@ -1539,6 +1604,18 @@ class GitHandler(object):
         if names:
             return names[0]
 
+    def audit_hg_path(self, path):
+        if '.hg' in path.split(os.path.sep):
+            if self.ui.configbool('git', 'blockdothg', True):
+                raise hgutil.Abort(
+                    ('Refusing to import problematic path %r' % path),
+                    hint=("Mercurial cannot check out paths inside nested " +
+                          "repositories; if you need to continue, then set " +
+                          "'[git] blockdothg = false' in your hgrc."))
+            self.ui.warn(('warning: path %r is within a nested repository, ' +
+                          'which Mercurial cannot check out.\n')
+                         % path)
+
     # Stolen from hgsubversion
     def swap_out_encoding(self, new_encoding='UTF-8'):
         try:
@@ -1568,9 +1645,12 @@ class GitHandler(object):
 
         Tests:
 
+        >>> from collections import namedtuple
         >>> from dulwich.client import HttpGitClient, SSHGitClient
         >>> from mercurial.ui import ui
-        >>> g = GitHandler('', ui())
+        >>> mockrepo = namedtuple('localrepo', ['vfs'])
+        >>> mockrepo.vfs = ''
+        >>> g = GitHandler(mockrepo, ui())
         >>> client, url = g.get_transport_and_path('http://fqdn.com/test.git')
         >>> print isinstance(client, HttpGitClient)
         True
@@ -1595,10 +1675,11 @@ class GitHandler(object):
         git_match = RE_GIT_URI.match(uri)
         if git_match:
             res = git_match.groupdict()
+            host, port, sepr = res['host'], res['port'], res['sepr']
             transport = client.TCPGitClient
             if 'ssh' in res['scheme']:
+                util.checksafessh(host)
                 transport = client.SSHGitClient
-            host, port, sepr = res['host'], res['port'], res['sepr']
             path = res['path']
             if sepr == '/' and not path.startswith('~'):
                 path = '/' + path
@@ -1615,7 +1696,9 @@ class GitHandler(object):
             uri = uri[4:]
 
         if uri.startswith('http://') or uri.startswith('https://'):
-            auth = urllib2.HTTPBasicAuthHandler(url.passwordmgr(self.ui))
+            pmgr = compat.passwordmgr(self.ui)
+            auth = urllib2.HTTPBasicAuthHandler(pmgr)
+
             opener = urllib2.build_opener(auth)
             ua = 'git/20x6 (hg-git ; uses dulwich and hg ; like git-core)'
             opener.addheaders = [('User-Agent', ua)]
